@@ -27,18 +27,15 @@
 #include "libxml.h"
 
 #include <string.h>
+#include <libxml/pattern.h>
 #include <libxml/xmlmemory.h>
 #include <libxml/tree.h>
-#include <libxml/hash.h>
 #include <libxml/dict.h>
 #include <libxml/xmlerror.h>
 #include <libxml/parserInternals.h>
-#include <libxml/pattern.h>
 
 #ifdef LIBXML_PATTERN_ENABLED
 
-/* #define DEBUG_STREAMING */
-
 #ifdef ERROR
 #undef ERROR
 #endif
@@ -347,9 +344,8 @@ xmlFreePatParserContext(xmlPatParserContextPtr ctxt) {
  * Returns -1 in case of failure, 0 otherwise.
  */
 static int
-xmlPatternAdd(xmlPatParserContextPtr ctxt ATTRIBUTE_UNUSED,
-                xmlPatternPtr comp,
-                xmlPatOp op, xmlChar * value, xmlChar * value2)
+xmlPatternAdd(xmlPatParserContextPtr ctxt, xmlPatternPtr comp,
+              xmlPatOp op, xmlChar * value, xmlChar * value2)
 {
     if (comp->nbStep >= comp->maxStep) {
         xmlStepOpPtr temp;
@@ -358,6 +354,7 @@ xmlPatternAdd(xmlPatParserContextPtr ctxt ATTRIBUTE_UNUSED,
         if (temp == NULL) {
 	    ERROR(ctxt, NULL, NULL,
 			     "xmlPatternAdd: realloc failed\n");
+            ctxt->error = -1;
 	    return (-1);
 	}
 	comp->steps = temp;
@@ -700,10 +697,6 @@ rollback:
  *									*
  ************************************************************************/
 
-#define TODO								\
-    xmlGenericError(xmlGenericErrorContext,				\
-	    "Unimplemented block at %s:%d\n",				\
-            __FILE__, __LINE__);
 #define CUR (*ctxt->cur)
 #define SKIP(val) ctxt->cur += (val)
 #define NXT(val) ctxt->cur[(val)]
@@ -720,14 +713,6 @@ rollback:
 #define PUSH(op, val, val2)						\
     if (xmlPatternAdd(ctxt, ctxt->comp, (op), (val), (val2))) goto error;
 
-#define XSLT_ERROR(X)							\
-    { xsltError(ctxt, __FILE__, __LINE__, X);				\
-      ctxt->error = (X); return; }
-
-#define XSLT_ERROR0(X)							\
-    { xsltError(ctxt, __FILE__, __LINE__, X);				\
-      ctxt->error = (X); return(0); }
-
 #if 0
 /**
  * xmlPatScanLiteral:
@@ -872,6 +857,8 @@ xmlPatScanNCName(xmlPatParserContextPtr ctxt) {
 	ret = (xmlChar *) xmlDictLookup(ctxt->dict, q, cur - q);
     else
 	ret = xmlStrndup(q, cur - q);
+    if (ret == NULL)
+        ctxt->error = -1;
     CUR_PTR = cur;
     return(ret);
 }
@@ -916,6 +903,8 @@ xmlCompileAttributeTest(xmlPatParserContextPtr ctxt) {
 
     SKIP_BLANKS;
     name = xmlPatScanNCName(ctxt);
+    if (ctxt->error < 0)
+        return;
     if (name == NULL) {
 	if (CUR == '*') {
 	    PUSH(XML_OP_ATTR, NULL, NULL);
@@ -935,7 +924,6 @@ xmlCompileAttributeTest(xmlPatParserContextPtr ctxt) {
 
 	if (IS_BLANK_CH(CUR)) {
 	    ERROR5(NULL, NULL, NULL, "Invalid QName.\n", NULL);
-	    XML_PAT_FREE_STRING(ctxt, prefix);
 	    ctxt->error = 1;
 	    goto error;
 	}
@@ -960,12 +948,12 @@ xmlCompileAttributeTest(xmlPatParserContextPtr ctxt) {
 		ERROR5(NULL, NULL, NULL,
 		    "xmlCompileAttributeTest : no namespace bound to prefix %s\n",
 		    prefix);
-	        XML_PAT_FREE_STRING(ctxt, prefix);
 		ctxt->error = 1;
 		goto error;
 	    }
 	}
-	XML_PAT_FREE_STRING(ctxt, prefix);
+        XML_PAT_FREE_STRING(ctxt, name);
+        name = NULL;
 	if (token == NULL) {
 	    if (CUR == '*') {
 		NEXT;
@@ -984,6 +972,8 @@ xmlCompileAttributeTest(xmlPatParserContextPtr ctxt) {
     }
     return;
 error:
+    if (name != NULL)
+	XML_PAT_FREE_STRING(ctxt, name);
     if (URL != NULL)
 	XML_PAT_FREE_STRING(ctxt, URL)
     if (token != NULL)
@@ -1034,6 +1024,8 @@ xmlCompileStepPattern(xmlPatParserContextPtr ctxt) {
 	return;
     }
     name = xmlPatScanNCName(ctxt);
+    if (ctxt->error < 0)
+        return;
     if (name == NULL) {
 	if (CUR == '*') {
 	    NEXT;
@@ -1250,6 +1242,8 @@ xmlCompilePathPattern(xmlPatParserContextPtr ctxt) {
     if (CUR == '@') {
 	NEXT;
 	xmlCompileAttributeTest(ctxt);
+        if (ctxt->error != 0)
+            goto error;
 	SKIP_BLANKS;
 	/* TODO: check for incompleteness */
 	if (CUR != 0) {
@@ -1415,62 +1409,6 @@ error_unfinished:
  *									*
  ************************************************************************/
 
-#ifdef DEBUG_STREAMING
-static void
-xmlDebugStreamComp(xmlStreamCompPtr stream) {
-    int i;
-
-    if (stream == NULL) {
-        printf("Stream: NULL\n");
-	return;
-    }
-    printf("Stream: %d steps\n", stream->nbStep);
-    for (i = 0;i < stream->nbStep;i++) {
-	if (stream->steps[i].ns != NULL) {
-	    printf("{%s}", stream->steps[i].ns);
-	}
-        if (stream->steps[i].name == NULL) {
-	    printf("* ");
-	} else {
-	    printf("%s ", stream->steps[i].name);
-	}
-	if (stream->steps[i].flags & XML_STREAM_STEP_ROOT)
-	    printf("root ");
-	if (stream->steps[i].flags & XML_STREAM_STEP_DESC)
-	    printf("// ");
-	if (stream->steps[i].flags & XML_STREAM_STEP_FINAL)
-	    printf("final ");
-	printf("\n");
-    }
-}
-static void
-xmlDebugStreamCtxt(xmlStreamCtxtPtr ctxt, int match) {
-    int i;
-
-    if (ctxt == NULL) {
-        printf("Stream: NULL\n");
-	return;
-    }
-    printf("Stream: level %d, %d states: ", ctxt->level, ctxt->nbState);
-    if (match)
-        printf("matches\n");
-    else
-        printf("\n");
-    for (i = 0;i < ctxt->nbState;i++) {
-        if (ctxt->states[2 * i] < 0)
-	    printf(" %d: free\n", i);
-	else {
-	    printf(" %d: step %d, level %d", i, ctxt->states[2 * i],
-	           ctxt->states[(2 * i) + 1]);
-            if (ctxt->comp->steps[ctxt->states[2 * i]].flags &
-	        XML_STREAM_STEP_DESC)
-	        printf(" //\n");
-	    else
-	        printf("\n");
-	}
-    }
-}
-#endif
 /**
  * xmlNewStreamComp:
  * @size: the number of expected steps
@@ -1729,9 +1667,6 @@ xmlStreamCompile(xmlPatternPtr comp) {
     stream->steps[s].flags |= XML_STREAM_STEP_FINAL;
     if (root)
 	stream->steps[0].flags |= XML_STREAM_STEP_ROOT;
-#ifdef DEBUG_STREAMING
-    xmlDebugStreamComp(stream);
-#endif
     comp->stream = stream;
     return(0);
 error:
@@ -1849,12 +1784,9 @@ static int
 xmlStreamPushInternal(xmlStreamCtxtPtr stream,
 		      const xmlChar *name, const xmlChar *ns,
 		      int nodeType) {
-    int ret = 0, err = 0, final = 0, tmp, i, m, match, stepNr, desc;
+    int ret = 0, final = 0, tmp, i, m, match, stepNr, desc;
     xmlStreamCompPtr comp;
     xmlStreamStep step;
-#ifdef DEBUG_STREAMING
-    xmlStreamCtxtPtr orig = stream;
-#endif
 
     if ((stream == NULL) || (stream->nbState < 0))
         return(-1);
@@ -1883,10 +1815,8 @@ xmlStreamPushInternal(xmlStreamCtxtPtr stream,
 			*/
 			ret = 1;
 		    } else if (comp->steps[0].flags & XML_STREAM_STEP_ROOT) {
-			/* TODO: Do we need this ? */
-			tmp = xmlStreamCtxtAddState(stream, 0, 0);
-			if (tmp < 0)
-			    err++;
+			if (xmlStreamCtxtAddState(stream, 0, 0) < 0)
+                            return(-1);
 		    }
 		}
 	    }
@@ -2039,22 +1969,12 @@ xmlStreamPushInternal(xmlStreamCtxtPtr stream,
 #endif /* if 0 ------------------------------------------------------- */
 	    if (match) {
 		final = step.flags & XML_STREAM_STEP_FINAL;
-		if (desc) {
-		    if (final) {
-			ret = 1;
-		    } else {
-			/* descending match create a new state */
-			xmlStreamCtxtAddState(stream, stepNr + 1,
-			                      stream->level + 1);
-		    }
-		} else {
-		    if (final) {
-			ret = 1;
-		    } else {
-			xmlStreamCtxtAddState(stream, stepNr + 1,
-			                      stream->level + 1);
-		    }
-		}
+                if (final) {
+                    ret = 1;
+                } else if (xmlStreamCtxtAddState(stream, stepNr + 1,
+                                                 stream->level + 1) < 0) {
+                    return(-1);
+                }
 		if ((ret != 1) && (step.flags & XML_STREAM_STEP_IN_SET)) {
 		    /*
 		    * Check if we have a special case like "foo/bar//.", where
@@ -2155,10 +2075,11 @@ compare:
 	}
 	final = step.flags & XML_STREAM_STEP_FINAL;
 	if (match) {
-	    if (final)
+	    if (final) {
 		ret = 1;
-	    else
-		xmlStreamCtxtAddState(stream, 1, stream->level);
+            } else if (xmlStreamCtxtAddState(stream, 1, stream->level) < 0) {
+                return(-1);
+            }
 	    if ((ret != 1) && (step.flags & XML_STREAM_STEP_IN_SET)) {
 		/*
 		* Check if we have a special case like "foo//.", where
@@ -2180,11 +2101,6 @@ stream_next:
         stream = stream->next;
     } /* while stream != NULL */
 
-    if (err > 0)
-        ret = -1;
-#ifdef DEBUG_STREAMING
-    xmlDebugStreamCtxt(orig, ret);
-#endif
     return(ret);
 }
 
@@ -2207,7 +2123,7 @@ stream_next:
 int
 xmlStreamPush(xmlStreamCtxtPtr stream,
               const xmlChar *name, const xmlChar *ns) {
-    return (xmlStreamPushInternal(stream, name, ns, (int) XML_ELEMENT_NODE));
+    return (xmlStreamPushInternal(stream, name, ns, XML_ELEMENT_NODE));
 }
 
 /**
@@ -2257,7 +2173,7 @@ xmlStreamPushNode(xmlStreamCtxtPtr stream,
 int
 xmlStreamPushAttr(xmlStreamCtxtPtr stream,
 		  const xmlChar *name, const xmlChar *ns) {
-    return (xmlStreamPushInternal(stream, name, ns, (int) XML_ATTRIBUTE_NODE));
+    return (xmlStreamPushInternal(stream, name, ns, XML_ATTRIBUTE_NODE));
 }
 
 /**
@@ -2336,28 +2252,37 @@ xmlStreamWantsAnyNode(xmlStreamCtxtPtr streamCtxt)
  ************************************************************************/
 
 /**
- * xmlPatterncompile:
+ * xmlPatternCompileSafe:
  * @pattern: the pattern to compile
  * @dict: an optional dictionary for interned strings
  * @flags: compilation flags, see xmlPatternFlags
  * @namespaces: the prefix definitions, array of [URI, prefix] or NULL
+ * @patternOut: output pattern
  *
  * Compile a pattern.
  *
- * Returns the compiled form of the pattern or NULL in case of error
+ * Available since 2.13.0.
+ *
+ * Returns 0 on success, 1 on error, -1 if a memory allocation failed.
  */
-xmlPatternPtr
-xmlPatterncompile(const xmlChar *pattern, xmlDict *dict, int flags,
-                  const xmlChar **namespaces) {
+int
+xmlPatternCompileSafe(const xmlChar *pattern, xmlDict *dict, int flags,
+                      const xmlChar **namespaces, xmlPatternPtr *patternOut) {
     xmlPatternPtr ret = NULL, cur;
     xmlPatParserContextPtr ctxt = NULL;
     const xmlChar *or, *start;
     xmlChar *tmp = NULL;
     int type = 0;
     int streamable = 1;
+    int error;
 
-    if (pattern == NULL)
-        return(NULL);
+    if (patternOut == NULL)
+        return(1);
+
+    if (pattern == NULL) {
+        error = 1;
+        goto error;
+    }
 
     start = pattern;
     or = start;
@@ -2373,9 +2298,15 @@ xmlPatterncompile(const xmlChar *pattern, xmlDict *dict, int flags,
 	    }
 	    or++;
 	}
-	if (ctxt == NULL) goto error;
+	if (ctxt == NULL) {
+            error = -1;
+            goto error;
+        }
 	cur = xmlNewPattern();
-	if (cur == NULL) goto error;
+	if (cur == NULL) {
+            error = -1;
+            goto error;
+        }
 	/*
 	* Assign string dict.
 	*/
@@ -2396,8 +2327,10 @@ xmlPatterncompile(const xmlChar *pattern, xmlDict *dict, int flags,
 	    xmlCompileIDCXPathPath(ctxt);
 	else
 	    xmlCompilePathPattern(ctxt);
-	if (ctxt->error != 0)
+	if (ctxt->error != 0) {
+            error = ctxt->error;
 	    goto error;
+        }
 	xmlFreePatParserContext(ctxt);
 	ctxt = NULL;
 
@@ -2413,9 +2346,13 @@ xmlPatterncompile(const xmlChar *pattern, xmlDict *dict, int flags,
 		    streamable = 0;
 	    }
 	}
-	if (streamable)
-	    xmlStreamCompile(cur);
-	if (xmlReversePattern(cur) < 0)
+	if (streamable) {
+	    error = xmlStreamCompile(cur);
+            if (error != 0)
+                goto error;
+        }
+	error = xmlReversePattern(cur);
+        if (error != 0)
 	    goto error;
 	if (tmp != NULL) {
 	    xmlFree(tmp);
@@ -2434,12 +2371,33 @@ xmlPatterncompile(const xmlChar *pattern, xmlDict *dict, int flags,
 	}
     }
 
-    return(ret);
+    *patternOut = ret;
+    return(0);
 error:
     if (ctxt != NULL) xmlFreePatParserContext(ctxt);
     if (ret != NULL) xmlFreePattern(ret);
     if (tmp != NULL) xmlFree(tmp);
-    return(NULL);
+    *patternOut = NULL;
+    return(error);
+}
+
+/**
+ * xmlPatterncompile:
+ * @pattern: the pattern to compile
+ * @dict: an optional dictionary for interned strings
+ * @flags: compilation flags, see xmlPatternFlags
+ * @namespaces: the prefix definitions, array of [URI, prefix] or NULL
+ *
+ * Compile a pattern.
+ *
+ * Returns the compiled form of the pattern or NULL in case of error
+ */
+xmlPatternPtr
+xmlPatterncompile(const xmlChar *pattern, xmlDict *dict, int flags,
+                  const xmlChar **namespaces) {
+    xmlPatternPtr ret;
+    xmlPatternCompileSafe(pattern, dict, flags, namespaces, &ret);
+    return(ret);
 }
 
 /**
